"""
Shared pytest fixtures for unit tests.
"""
import numpy as np
import pytest


@pytest.fixture
def data_vandermonde():
    r = 11
    n = 200
    x = np.linspace(0, 1, n + 1)
    vde = np.vander(x, r, increasing=True)

    # PySensor objects expect rows to correspond to examples,
    # columns to positions
    return vde.T


# Same as above, but also returns test data
@pytest.fixture
def data_vandermonde_testing():
    r = 11
    n = 200
    x = np.linspace(0, 1, n + 1)
    vde = np.zeros((n + 1, r))
    vde[:, 0] = np.ones(n + 1)

    for i in range(r - 1):
        vde[:, i + 1] = vde[:, i] * x

    v = np.zeros(r)
    v[[1, 3, 5]] = 1
    x_test = np.dot(vde, v)

    # PySensor objects expect rows to correspond to examples,
    # columns to positions
    return vde.T, x_test


@pytest.fixture
def data_random():
    n_examples = 30
    n_features = 20

    return np.random.randn(n_examples, n_features)


@pytest.fixture
def data_random_square():
    n_examples = 30
    n_features = 30

    return np.random.randn(n_examples, n_features)



from abc import ABC
from abc import abstractmethod

from sklearn.utils.validation import check_is_fitted


class InvertibleBasis(ABC):
    """
    An abstract class ensuring subclasses implement a
    ``matrix_inverse`` method.
    """

    @abstractmethod
    def matrix_inverse(self, n_basis_modes=None, **kwargs):
        raise NotImplementedError("This method has not been implemented")


class MatrixMixin:
    """
    Mixin class for generating matrix representations of a basis.
    """

    def matrix_representation(self, n_basis_modes=None, copy=False):
        """
        Get the matrix representation of the training data in the basis.
        Note that in general this matrix is not the matrix whose column vectors
        are the basis modes.

        Parameters
        ----------
        n_basis_modes: positive int, optional (default None)
            Number of basis modes of matrix representation to return.

        copy : boolean, optional (default False)
            Whether to return a copy of the basis matrix.

        Returns
        -------
        B : numpy array, shape (n_features, n_basis_modes)
            Matrix representation of the basis. Note that rows correspond to
            features and columns to basis modes.
        """
        n_basis_modes = self._validate_input(n_basis_modes)

        if copy:
            return self.basis_matrix_[:, :n_basis_modes].copy()
        else:
            return self.basis_matrix_[:, :n_basis_modes]

    def _validate_input(self, n_basis_modes):
        """
        Ensure ``n_basis_modes`` does not exceed the maximum number possible.
        """
        check_is_fitted(self, "basis_matrix_")

        if n_basis_modes is None:
            n_basis_modes = self.n_basis_modes
        elif n_basis_modes > self.n_basis_modes:
            raise ValueError(
                f"Requested number of modes {n_basis_modes} exceeds"
                f" number available: {self.n_basis_modes}"
            )
        return n_basis_modes



from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

try:
    __version__ = get_distribution("python-sensors").version
except DistributionNotFound:
    pass

from .classification import SSPOC
from .reconstruction import SSPOR


__all__ = [
    # Modules:
    "basis",
    "classification",
    "reconstruction",
    "optimizers",
    "utils",
    # Non-modules:
    "SSPOR",
    "SSPOC",
]



"""
Unit tests for SSPOR class (and any other classes implemented
in pysensors.py).

Note: all tests should be encapsulated in functions whose
names start with "test_"

To run all tests for this package, navigate to the top-level
directory and execute the following command:
pytest

To run tests for just one file, run
pytest file_to_test.py
"""
import pytest
from numpy import isnan
from numpy import mean
from numpy import nan
from numpy import sqrt
from numpy import zeros
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from pysensors.basis import Identity
from pysensors.basis import RandomProjection
from pysensors.basis import SVD
from pysensors.optimizers import CCQR
from pysensors.reconstruction import SSPOR


def test_not_fitted(data_vandermonde):
    x = data_vandermonde
    model = SSPOR()

    # Should not be able to call any of these methods before fitting
    with pytest.raises(NotFittedError):
        model.predict(x)
    with pytest.raises(NotFittedError):
        model.get_selected_sensors()
    with pytest.raises(NotFittedError):
        model.get_all_sensors()
    with pytest.raises(NotFittedError):
        model.set_number_of_sensors(20)
    with pytest.raises(NotFittedError):
        model.score(x)
    with pytest.raises(NotFittedError):
        model.reconstruction_error(x)


def test_set_number_of_sensors(data_vandermonde):
    x = data_vandermonde
    max_sensors = x.shape[1]

    model = SSPOR()
    model.fit(x)

    with pytest.raises(ValueError):
        model.set_number_of_sensors(max_sensors + 1)
    with pytest.raises(ValueError):
        model.set_number_of_sensors(0)
    with pytest.raises(ValueError):
        model.set_number_of_sensors(1.5)
    with pytest.raises(ValueError):
        model.set_number_of_sensors("3")

    model.set_number_of_sensors(15)
    assert len(model.get_selected_sensors()) == 15


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_vandermonde"), pytest.lazy_fixture("data_random")],
)
def test_get_all_sensors(data):
    x = data
    max_sensors = x.shape[1]

    model = SSPOR()
    model.fit(x)
    assert len(model.get_all_sensors()) == max_sensors


@pytest.mark.parametrize(
    "basis", [Identity(), SVD(), RandomProjection(n_basis_modes=5)]
)
def test_basis_compatibility(data_vandermonde, basis):
    x = data_vandermonde
    model = SSPOR(basis=basis)
    model.fit(x)
    check_is_fitted(model)


def test_n_sensors(data_random):

    # Check for bad inputs
    with pytest.raises(ValueError):
        model = SSPOR(n_sensors=0)
    with pytest.raises(ValueError):
        model = SSPOR(n_sensors=5.4)
    with pytest.raises(ValueError):
        model = SSPOR(n_sensors="1")
    with pytest.raises(ValueError):
        model = SSPOR(n_sensors=[1])

    n_sensors = 5
    x = data_random
    model = SSPOR(n_sensors=n_sensors)
    model.fit(x)

    assert len(model.get_selected_sensors()) == n_sensors


def test_predict(data_random):
    data = data_random

    n_sensors = 5
    model = SSPOR(n_sensors=n_sensors)
    model.fit(data)

    # Wrong size input for predict
    # (should only pass data at sensor locations)
    with pytest.raises(ValueError):
        model.predict(data)

    # Rectangular case
    sensors = model.get_selected_sensors()
    assert data.shape == model.predict(data[:, sensors]).shape


def test_square_predict(data_random_square):
    data = data_random_square

    model = SSPOR()
    model.fit(data)
    sensors = model.get_selected_sensors()
    assert data.shape == model.predict(data[:, sensors]).shape


def test_predict_accuracy(data_vandermonde_testing):
    # Polynomials up to degree 10 on [0, 1]
    data, x_test = data_vandermonde_testing

    model = SSPOR()
    model.fit(data, seed=1)
    model.set_number_of_sensors(8)
    sensors = model.get_selected_sensors()
    assert sqrt(mean((x_test - model.predict(x_test[sensors])) ** 2)) <= 1.0e-3

    # Should also work for row vectors
    x_test = x_test.reshape(1, -1)
    assert sqrt(mean((x_test - model.predict(x_test[:, sensors])) ** 2)) <= 1.0e-3


def test_reconstruction_error(data_vandermonde_testing):
    data, x_test = data_vandermonde_testing

    model = SSPOR(n_sensors=3)
    model.fit(data)

    assert len(model.reconstruction_error(x_test)) == min(
        model.n_sensors, data.shape[1]
    )

    sensor_range = [1, 2, 3]
    assert len(model.reconstruction_error(x_test, sensor_range=sensor_range)) == 3


def test_score(data_vandermonde):
    data = data_vandermonde

    weak_model = SSPOR(n_sensors=3)
    weak_model.fit(data)

    # You must pass in data with as many features as the training set
    with pytest.raises(ValueError):
        weak_model.score(data[:, :5])

    strong_model = SSPOR(n_sensors=8)
    strong_model.fit(data)

    assert weak_model.score(data) < strong_model.score(data)


def test_prefit_basis(data_random):
    data = data_random
    basis = Identity()
    basis.fit(data)

    # This data should be ignored during the fit
    data_to_ignore = nan * data_random

    model = SSPOR(basis=basis)
    model.fit(data_to_ignore, prefit_basis=True)
    assert not any(isnan(model.get_selected_sensors()))


def test_update_n_basis_modes_errors(data_random):
    data = data_random
    n_basis_modes = 5
    model = SSPOR(basis=Identity(n_basis_modes=n_basis_modes))

    model.fit(data)

    with pytest.raises(ValueError):
        model.update_n_basis_modes(0)
    with pytest.raises(ValueError):
        model.update_n_basis_modes("5")
    with pytest.raises(ValueError):
        model.update_n_basis_modes(data.shape[0] + 1)
    # Need to pass x when increasing n_basis_modes beyond capacity
    # of the original basis
    with pytest.raises(ValueError):
        model.update_n_basis_modes(n_basis_modes + 1)


def test_update_n_basis_modes(data_random):
    data = data_random
    model = SSPOR()
    model.fit(data)
    assert model.basis.n_basis_modes == data.shape[0]
    assert model.basis_matrix_.shape[1] == data.shape[0]

    n_basis_modes = 5
    model.update_n_basis_modes(n_basis_modes)
    assert model.basis.n_basis_modes == data.shape[0]
    assert model.basis_matrix_.shape[1] == n_basis_modes


def test_update_n_basis_modes_refit(data_random):
    data = data_random
    n_basis_modes = 5
    model = SSPOR(basis=Identity(n_basis_modes=n_basis_modes))
    model.fit(data)
    assert model.basis_matrix_.shape[1] == n_basis_modes

    model.update_n_basis_modes(n_basis_modes + 1, data)
    assert model.basis_matrix_.shape[1] == n_basis_modes + 1


def test_update_n_basis_modes_unfit_basis(data_random):
    data = data_random
    n_basis_modes = 5
    model = SSPOR()
    model.update_n_basis_modes(n_basis_modes, data)

    assert model.basis_matrix_.shape[1] == n_basis_modes


def test_ccqr_integration(data_random):
    data = data_random
    costs = zeros(data.shape[1])
    costs[[1, 3, 5]] = 100

    optimizer = CCQR(sensor_costs=costs)
    model = SSPOR(optimizer=optimizer).fit(data)

    check_is_fitted(model)


def test_sensor_selector_properties(data_random):
    data = data_random
    model = SSPOR().fit(data)

    assert all(model.get_all_sensors() == model.all_sensors)
    assert all(model.get_selected_sensors() == model.selected_sensors)



"""
Various utility functions.
"""
import numpy as np

# from sklearn.utils.validation import check_array


def validate_input(x, sensors=None):
    """
    Ensure that x is of compatible type and shape.

    Parameters
    ----------

    x: numpy ndarray, shape [n_features,] or [n_examples, n_features]
        Data to be validated.
    """
    # TODO: implement more checks
    if not isinstance(x, np.ndarray):
        raise ValueError("x must be a numpy array")

    # check_array(x)

    if sensors is not None:
        n_features = len(x) if np.ndim(x) == 1 else x.shape[1]
        if len(sensors) != n_features:
            raise ValueError(
                """x has the wrong number of features: {}.
                Expected {}""".format(
                    n_features, len(sensors)
                )
            )

    return x



from ._base import validate_input
from ._optimizers import constrained_binary_solve
from ._optimizers import constrained_multiclass_solve


__all__ = [
    "constrained_binary_solve",
    "constrained_multiclass_solve",
    "validate_input",
]



import warnings

import numpy as np
from scipy.linalg import lstsq
from scipy.linalg import solve
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted

from ..basis import Identity
from ..optimizers import QR
from ..utils import validate_input


INT_DTYPES = (int, np.int64, np.int32, np.int16, np.int8)


class SSPOR(BaseEstimator):
    """
    Sparse Sensor Placement Optimization for Reconstruction:
    a model for selecting the best sensor locations for state reconstruction.

    Given a basis in which to represent the state (e.g. PCA modes) along with
    measurement data, a :class:`SSPOR` instance produces a list of
    sensor locations (a permutation of the numbers 0, 1, ...,
    :code:`n_input_features` - 1) ranked in descending order of importance.
    One can then select the top k sensors and take future measurements at
    that limited set of locations.

    The overall time complexity of fitting a :class:`SSPOR` object is
    :code:`O(n_basis_modes * n_input_features * n_input_features)`
    plus the cost for fitting the basis. Different bases have different
    complexities. The space complexity is
    :code:`O(n_basis_modes * n_input_features)`.

    Parameters
    ----------
    basis: basis object, optional (default :class:`pysensors.basis.Identity`)
        Basis in which to represent the data. Default is the identity basis
        (i.e. raw features).

    optimizer: optimizer object, optional \
            (default :class:`pysensors.optimizers.QR`)
        Optimization method used to rank sensor locations.

    n_sensors: int, optional (default n_input_features)
        Number of sensors to select. Note that
        ``s = SSPOR(n_sensors=10); s.fit(x)``
        is equivalent to
        ``s = SSPOR(); s.fit(x); s.set_number_of_sensors(10)``.

    Attributes
    ----------
    n_basis_modes: int
        Number of basis modes considered during fitting.

    basis_matrix_: np.ndarray
        Internal representation of the basis.

    ranked_sensors_: np.ndarray
        Sensor locations ranked in descending order of importance.

    Examples
    --------
    >>> import numpy as np
    >>> from pysensors import SSPOR
    >>>
    >>> x = np.linspace(0, 1, 501)
    >>> monomials = np.vander(x, 15).T
    >>>
    >>> model = SSPOR(n_sensors=5)
    >>> model.fit(monomials)
    SSPOR(basis=Identity(n_basis_modes=15), n_sensors=5, optimizer=QR())
    >>> print(model.selected_sensors)
    [500 377   0 460 185]
    >>> print(x[model.selected_sensors])
    [1.    0.754 0.    0.92  0.37 ]
    >>> model.set_n_sensors(7)
    >>> print(x[model.selected_sensors])
    [1.    0.754 0.    0.92  0.37  0.572 0.134]
    >>> f = np.sin(3*x)
    >>> f_pred = model.predict(f[model.selected_sensors])
    >>> print(np.linalg.norm(f - f_pred))
    0.022405698005838044
    """

    def __init__(self, basis=None, optimizer=None, n_sensors=None):
        if basis is None:
            basis = Identity()
        self.basis = basis
        if optimizer is None:
            optimizer = QR()
        self.optimizer = optimizer
        if n_sensors is None:
            self.n_sensors = None
        elif isinstance(n_sensors, INT_DTYPES) and n_sensors > 0:
            self.n_sensors = int(n_sensors)
        else:
            raise ValueError("n_sensors must be a positive integer.")
        self.n_basis_modes = None

    def fit(self, x, quiet=False, prefit_basis=False, seed=None, **optimizer_kws):
        """
        Fit the SSPOR model, determining which sensors are relevant.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Training data.

        quiet: boolean, optional (default False)
            Whether or not to suppress warnings during fitting.

        prefit_basis: boolean, optional (default False)
            Whether or not the basis has already been fit to x.
            For example, you may have already fit and experimented with
            a ``SVD`` object to determine the optimal number of modes. This
            option allows you to avoid an unnecessary SVD.

        seed: int, optional (default None)
            Seed for the random number generator used to shuffle sensors after the
            ``self.basis.n_basis_modes`` sensor. Most optimizers only rank the top
            ``self.basis.n_basis_modes`` sensors, leaving the rest virtually
            untouched. As a result the remaining samples are randomly permuted.

        optimizer_kws: dict, optional
            Keyword arguments to be passed to the ``get_sensors`` method of
            the optimizer.

        Returns
        -------
        self: a fitted :class:`SSPOR` instance
        """

        # Fit basis functions to data
        if prefit_basis:
            check_is_fitted(self.basis, "basis_matrix_")
        else:
            x = validate_input(x)

            if quiet:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    self.basis.fit(x)
            else:
                self.basis.fit(x)

        # Get matrix representation of basis
        self.basis_matrix_ = self.basis.matrix_representation(
            n_basis_modes=self.n_basis_modes
        )

        # Check that n_sensors doesn't exceed dimension of basis vectors
        self._validate_n_sensors()

        # Find sparse sensor locations
        self.ranked_sensors_ = self.optimizer.fit(
            self.basis_matrix_, **optimizer_kws
        ).get_sensors()

        # Randomly shuffle sensors after self.basis.n_basis_modes
        rng = np.random.default_rng(seed)
        n_basis_modes = self.basis_matrix_.shape[1]
        self.ranked_sensors_[n_basis_modes:] = rng.permutation(
            self.ranked_sensors_[n_basis_modes:]
        )

        return self

    def predict(self, x, **solve_kws):
        """
        Predict values at all positions given measurements at sensor locations.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_sensors)
            Measurements from which to form prediction.
            The measurements should be taken at the sensor locations specified by
            ``self.get_selected_sensors()``.

        solve_kws: dict, optional
            keyword arguments to be passed to the linear solver used to invert
            the basis matrix.

        Returns
        -------
        y: numpy array, shape (n_samples, n_features)
            Predicted values at every location.
        """
        check_is_fitted(self, "ranked_sensors_")
        x = validate_input(x, self.ranked_sensors_[: self.n_sensors]).T

        # For efficiency we may want to factor
        # self.basis_matrix_[self.ranked_sensors_, :]
        # in case predict is called multiple times.
        # Although if the user changes the number of sensors between calls
        # the factorization will be wasted.

        if self.n_sensors > self.basis_matrix_.shape[0]:
            warnings.warn(
                "n_sensors exceeds dimension of basis modes. Performance may be poor"
            )

        # Square matrix
        if self.n_sensors == self.basis_matrix_.shape[1]:
            return self._square_predict(
                x, self.ranked_sensors_[: self.n_sensors], **solve_kws
            )
        # Rectangular matrix
        else:
            return self._rectangular_predict(
                x, self.ranked_sensors_[: self.n_sensors], **solve_kws
            )

    def _square_predict(self, x, sensors, **solve_kws):
        """Get prediction when the problem is square."""
        return np.dot(
            self.basis_matrix_, solve(self.basis_matrix_[sensors, :], x, **solve_kws)
        ).T

    def _rectangular_predict(self, x, sensors, **solve_kws):
        """Get prediction when the problem is rectangular."""
        return np.dot(
            self.basis_matrix_, lstsq(self.basis_matrix_[sensors, :], x, **solve_kws)[0]
        ).T

    def get_selected_sensors(self):
        """
        Get the indices of the sensors chosen by the model.

        Returns
        -------
        sensors: numpy array, shape (n_sensors,)
            Indices of the sensors chosen by the model
            (i.e. the sensor locations) ranked in descending order
            of importance.
        """
        check_is_fitted(self, "ranked_sensors_")
        return self.ranked_sensors_[: self.n_sensors]

    @property
    def selected_sensors(self):
        """
        Get the indices of the sensors chosen by the model.

        Returns
        -------
        sensors: numpy array, shape (n_sensors,)
            Indices of the sensors chosen by the model
            (i.e. the sensor locations) ranked in descending order
            of importance.
        """
        return self.get_selected_sensors()

    def get_all_sensors(self):
        """
        Get a ranked list consisting of all the sensors.
        The sensors are given in descending order of importance.

        Returns
        -------
        sensors: numpy array, shape (n_features,)
            Indices of sensors in descending order of importance.
        """
        return self.all_sensors

    @property
    def all_sensors(self):
        """
        Get a ranked list consisting of all the sensors.
        The sensors are given in descending order of importance.

        Returns
        -------
        sensors: numpy array, shape (n_features,)
            Indices of sensors in descending order of importance.
        """
        check_is_fitted(self, "ranked_sensors_")
        return self.ranked_sensors_

    def set_number_of_sensors(self, n_sensors):
        """
        Set ``n_sensors``, the number of sensors to be used for prediction.

        Parameters
        ----------
        n_sensors: int
            The number of sensors. Must be a positive integer.
            Cannot exceed the number of available sensors (n_features).
        """
        check_is_fitted(self, "ranked_sensors_")

        if not isinstance(n_sensors, INT_DTYPES):
            raise ValueError("n_sensors must be a positive integer")
        elif n_sensors <= 0:
            raise ValueError("n_sensors must be a positive integer")
        elif n_sensors > len(self.ranked_sensors_):
            raise ValueError(
                "n_sensors cannot exceed number of available sensors: "
                "{}".format(len(self.ranked_sensors_))
            )
        else:
            self.n_sensors = n_sensors

    def set_n_sensors(self, n_sensors):
        """
        A convenience function accomplishing the same thing as
        :meth:`set_number_of_sensors`.
        Set ``n_sensors``, the number of sensors to be used for prediction.

        Parameters
        ----------
        n_sensors: int
            The number of sensors. Must be a positive integer.
            Cannot exceed the number of available sensors (n_features).
        """
        self.set_number_of_sensors(n_sensors)

    def update_n_basis_modes(self, n_basis_modes, x=None, quiet=False):
        """
        Re-fit the :class:`SSPOR` object using a different value of
        ``n_basis_modes``.

        This method allows one to relearn sensor locations for a
        different number of basis modes _without_ re-fitting the basis
        in many cases.
        Specifically, if ``n_basis_modes <= self.basis.n_basis_modes``
        then the basis does not need to be refit.
        Otherwise this function does not save any computational resources.

        Parameters
        ----------
        n_basis_modes: positive int, optional (default None)
            Number of basis modes to be used during fit.
            Must be less than or equal to ``n_samples``.

        x: numpy array, shape (n_examples, n_features), optional (default None)
            Only used if ``n_basis_modes`` exceeds the number of available
            basis modes for the already fit basis.

        quiet: boolean, optional (default False)
            Whether or not to suppress warnings during refitting.
        """
        if not isinstance(n_basis_modes, INT_DTYPES) or n_basis_modes <= 0:
            raise ValueError("n_basis_modes must be a positive integer")

        # No need to refit basis; only refit sensors
        if (
            hasattr(self.basis, "basis_matrix_")
            and n_basis_modes <= self.basis.n_basis_modes
        ):
            self.n_basis_modes = n_basis_modes
            self.fit(x, prefit_basis=True, quiet=quiet)

        elif x is None:
            raise ValueError(
                "x cannot be None when n_basis_modes exceeds number of available modes"
            )
        elif n_basis_modes > x.shape[0]:
            raise ValueError(
                "n_basis_modes cannot exceed the number of examples, x.shape[0]"
            )
        else:
            self.n_basis_modes = n_basis_modes
            self.basis.n_basis_modes = n_basis_modes
            self.fit(x, prefit_basis=False, quiet=quiet)

    def score(self, x, y=None, score_function=None, score_kws={}, solve_kws={}):
        """
        Compute the reconstruction error for a given set of measurements.

        Parameters
        ----------
        x: numpy array, shape (n_examples, n_features)
            Measurements with which to compute the score.
            Note that ``x`` should consist of measurements at every location,
            not just the recommended sensor location, i.e. its shape should be
            (n_examples, n_features) rather than (n_examples, n_sensors).

        y: None
            Dummy input to maintain compatibility with Scikit-learn.

        score_function: callable, optional (default None)
            Function used to compute the score. Should have the call signature
            ``score_function(y_true, y_pred, **score_kws)``.
            Default is the negative of the root mean squared error
            (sklearn expects higher scores to correspond to better performance).

        score_kws: dict, optional
            Keyword arguments to be passed to score_function. Ignored if
            score_function is None.

        solve_kws: dict, optional
            Keyword arguments to be passed to the predict method.

        Returns
        -------
        score: float
            The score.
        """
        check_is_fitted(self, "ranked_sensors_")

        n_input_features = len(x) if np.ndim(x) == 1 else x.shape[1]
        n_expected_features = len(self.ranked_sensors_)
        if n_expected_features != n_input_features:
            raise ValueError(
                f"x has {n_input_features} features (columns), "
                f"but should have {n_expected_features}"
            )

        sensors = self.get_selected_sensors()
        if score_function is None:
            return -np.sqrt(
                np.mean((self.predict(x[:, sensors], **solve_kws) - x) ** 2)
            )
        else:
            return score_function(
                x,
                self.predict(x[:, sensors], **solve_kws),
                **score_kws,
            )

    def reconstruction_error(self, x_test, sensor_range=None, score=None, **solve_kws):
        """
        Compute the reconstruction error for different numbers of sensors.

        Parameters
        ----------
        x_test: numpy array, shape (n_examples, n_features)
            Measurements to be reconstructed.

        sensor_range: 1D numpy array, optional (default None)
            Numbers of sensors at which to compute the reconstruction error.
            If None, will be set to
            [1, 2, ... , min(``n_sensors``, ``basis.n_basis_modes``)].

        score: callable, optional (default None)
            Function used to compute the reconstruction error.
            Should have the signature ``score(x, x_pred)``.
            If None, the root mean squared error is used.

        solve_kws: dict, optional
            Keyword arguments to be passed to the linear solver.

        Returns
        -------
        error: numpy array, shape (len(sensor_range),)
            Reconstruction scores for each number of sensors in ``sensor_range``.
        """
        check_is_fitted(self, "ranked_sensors_")
        x_test = validate_input(x_test, self.get_all_sensors()).T

        basis_mode_dim, n_basis_modes = self.basis_matrix_.shape
        if sensor_range is None:
            sensor_range = np.arange(1, min(self.n_sensors, basis_mode_dim) + 1)
        if sensor_range[-1] > basis_mode_dim:
            warnings.warn(
                f"Performance may be poor when using more than {basis_mode_dim} sensors"
            )

        if score is None:

            def score(x, y):
                return np.sqrt(np.mean((x - y) ** 2))

        error = np.zeros_like(sensor_range, dtype=np.float64)

        for k, n_sensors in enumerate(sensor_range):
            if n_sensors == n_basis_modes:
                error[k] = score(
                    self._square_predict(
                        x_test[self.ranked_sensors_[:n_sensors]],
                        self.ranked_sensors_[:n_sensors],
                        **solve_kws,
                    ),
                    x_test.T,
                )
            else:
                error[k] = score(
                    self._rectangular_predict(
                        x_test[self.ranked_sensors_[:n_sensors]],
                        self.ranked_sensors_[:n_sensors],
                        **solve_kws,
                    ),
                    x_test.T,
                )

        return error

    def _validate_n_sensors(self):
        """
        Check that number of sensors does not exceed the maximimum number
        allowed by the chosen basis.
        """
        check_is_fitted(self, "basis_matrix_")

        # Maximum number of sensors (= dimension of basis vectors)
        max_sensors = self.basis_matrix_.shape[0]
        if self.n_sensors is None:
            self.n_sensors = max_sensors
        elif self.n_sensors > max_sensors:
            raise ValueError(
                "n_sensors cannot exceed number of available sensors: {}".format(
                    max_sensors
                )
            )



"""
Identity basis class.

This is essentially a dummy basis which just uses raw, unaltered features.
"""
from warnings import warn

from numpy import identity
from sklearn.base import BaseEstimator
from sklearn.utils import check_array

from ._base import InvertibleBasis
from ._base import MatrixMixin


class Identity(BaseEstimator, InvertibleBasis, MatrixMixin):
    """
    Generate an identity transformation which maps all input features to
    themselves.

    Parameters
    ----------
    n_basis_modes: int, optional (default None)
        Number of basis modes to retain. If None, all are included
        in the basis.

    Attributes
    ----------
    basis_matrix_ : numpy ndarray, shape (n_features, n_basis_modes)
        The transpose of the first ``n_basis_modes`` examples from the
        input data.
    """

    def __init__(self, n_basis_modes=None):
        if n_basis_modes is None:
            self.n_basis_modes = None
        elif isinstance(n_basis_modes, int) and n_basis_modes > 0:
            self.n_basis_modes = n_basis_modes
        else:
            raise ValueError("n_basis_modes must be a positive integer.")

    def fit(self, X):
        """
        Memorize the input data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training data.

        Returns
        -------
        self : instance
        """

        # Note that we take a transpose here, so columns correspond to examples
        if self.n_basis_modes is None:
            self.basis_matrix_ = check_array(X).T.copy()
            self.n_basis_modes = self.basis_matrix_.shape[1]
        else:
            if self.n_basis_modes > X.shape[0]:
                raise ValueError(
                    "X needs at least n_basis_modes ({}) examples/rows".format(
                        self.n_basis_modes
                    )
                )

            self.basis_matrix_ = check_array(X)[: self.n_basis_modes, :].T.copy()

            if self.n_basis_modes < X.shape[0]:
                warn(f"Only the first {self.n_basis_modes} examples were retained.")
        return self

    def matrix_inverse(self, n_basis_modes=None):
        """
        Get the inverse matrix mapping from measurement space to
        coordinates with respect to the basis.

        Note that this is not the inverse of the matrix returned by
        ``self.matrix_representation``. It is the (psuedo) inverse of
        the matrix whose columns are the basis modes.

        Parameters
        ----------
        n_basis_modes : positive int, optional (default None)
            Number of basis modes to be used to compute inverse.

        Returns
        -------
        B : numpy ndarray, shape (n_features, n_features)
            The inverse matrix. In this case B is the identity matrix.
        """
        # TODO: validate this
        n_basis_modes = self._validate_input(n_basis_modes)

        return identity(self.basis_matrix_.shape[0])



from scipy.linalg import qr
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted


class QR(BaseEstimator):
    """
    Greedy QR optimizer for sensor selection.
    Ranks sensors in descending order of "importance" by applying
    the QR algorithm and extracting pivot indices.

    See the following reference for more information

        Manohar, Krithika, et al.
        "Data-driven sparse sensor placement for reconstruction:
        Demonstrating the benefits of exploiting known patterns."
        IEEE Control Systems Magazine 38.3 (2018): 63-86.
    """

    def __init__(self):
        """
        Attributes
        ----------
        pivots_ : np.ndarray, shape [n_features]
            Ranked list of sensor locations.
        """
        self.pivots_ = None

    def fit(self, basis_matrix, **optimizer_kws):
        """
        Parameters
        ----------
        basis_matrix: np.ndarray, shape [n_features, n_samples]
            Matrix whose columns are the basis vectors in which to
            represent the measurement data.

        optimizer_kws: dictionary, optional
            Keyword arguments to be passed to the qr method.

        Returns
        -------

        """

        # TODO: implement checks on basis_matrix
        _, _, self.pivots_ = qr(basis_matrix.conj().T, pivoting=True, **optimizer_kws)

        return self

    def get_sensors(self):
        """
        Get ranked array of sensors.

        Returns
        -------
        sensors: np.ndarray, shape [n_features,]
            Array of sensors ranked in descending order of importance.
            Note that if n_features exceeds n_samples, then only the first
            n_samples entries of sensors are guaranteed to be in ranked order.
        """
        check_is_fitted(self, "pivots_")
        return self.pivots_
